/*
 * classMethod.c
 * Dictionary of classes, methods and fields.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define DBG(s)
#define	CDBG(s)
#define	MDBG(s)
#define	FDBG(s)
#define	LDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "slots.h"
#include "access.h"
#include "object.h"
#include "classMethod.h"
#include "code.h"
#include "file.h"
#include "readClass.h"
#include "baseClasses.h"
#include "errors.h"
#include "exception.h"
#include "external.h"
#include "lookup.h"
#include "support.h"

#define	CLASSHASHSZ		256
#if CLASSHASHSZ & (CLASSHASHSZ - 1)
  #error CLASSHASHSZ must be power of two
#endif
static classes* classPool[CLASSHASHSZ];
methods* methodList;	/* For exception processing */

extern void linkClass (classes*);
extern classes* findClass(char*);
static void prepareClass (classes*);
static classes*	simpleLookupClass(Utf8Const*, struct _classLoader*);
static classes* internalAddClass(Class*, Utf8Const*, int, int, struct _classLoader*);

classes*
addClass(Class* cl, constIndex c, constIndex s, u2 flags,
	 struct _classLoader* loader)
{
	Utf8Const* str;
	constants* pool = CLASS_CONSTANTS (cl);

	/* Find the name of the class */
	if (pool->tags[c] != CONSTANT_Class) {
CDBG(		printf("addClass: not a class.\n");			)
		return (0);
	}
	c = CLASS_CONST_USHORT1(cl, c);
	if (pool->tags[c] != CONSTANT_Utf8) {
CDBG(		printf("addClass: not a Utf8.\n");			)
		return (0);
	}
	str = WORD2UTF(pool->data[c]);

	return (internalAddClass(cl, str, flags, s, loader));
}

#if INTERN_UTF8CONSTS
static
int
hashClassName (Utf8Const *name)
{
	int len = (name->length+1) >> 1;  /* Number of shorts */
	uint16 *ptr = (uint16*) name->data;
	int hash = len;
	while (--len >= 0)
		hash = (hash << 1) ^ *ptr++;
	return hash;
}
#else
#define hashClassName(NAME) ((NAME)->hash)
#endif

static
classes*
internalAddClass(Class* cl, Utf8Const* name, int flags, int su,
		struct _classLoader* loader)
{
	Class** clp;
	uint32 hash = hashClassName (name) & (CLASSHASHSZ-1);
	int i;

	/* All classes are entered in the global classPool.
	 * This is done, even for classes using a ClassLoader, following
	 * the prescription of the "Java Language Specification"
	 * (section 12.2:  Loading of Classes and Interfaces). */

	clp = &classPool[hash];
#ifdef DEBUG
	while (*clp != 0) {
		assert(! equalUtf8Consts (name, (*clp)->name)
		       || loader != (*clp)->loader);
		clp = &(*clp)->next;
	}
#endif

	cl->name = name;
	cl->methodList = 0;
	cl->fieldList = 0;
	cl->staticFieldList = 0;
	cl->superclass = (Class*) (size_t) su;
	cl->msize = 0;
	cl->fsize = 0;
	cl->sfsize = 0;
	cl->accflags = flags;
	cl->dtable = 0;
	cl->staticFields = 0;
        cl->interfaces = 0;
	cl->interface_len = 0;
	cl->state = CSTATE_LOADED;
	cl->final = false;
	cl->loader = loader;
	/* Add into list. */
	cl->next = *clp;
	*clp = cl;

	return (cl);
}

Method*
addMethod(Class* c, method_info* m)
{
	constIndex nc;
	constIndex sc;
	methods** mptr;
	methods* mt;
	constants* pool;
	Utf8Const* name;
	Utf8Const* signature;
	Class* spc;

	pool = CLASS_CONSTANTS (c);

	nc = m->name_index;
	if (pool->tags[nc] != CONSTANT_Utf8) {
MDBG(		printf("addMethod: no method name.\n");			)
		return (0);
	}
	sc = m->signature_index;
	if (pool->tags[sc] != CONSTANT_Utf8) {
MDBG(		printf("addMethod: no signature name.\n");		)
		return (0);
	}
	name = WORD2UTF (pool->data[nc]);
	signature = WORD2UTF (pool->data[sc]);

	if (equalUtf8Consts (name, final_name)
	    && equalUtf8Consts (signature, void_signature)) {
		c->final = true;
	}

	mptr = &c->methodList;
#ifdef DEBUG
	/* Search down class for method name - don't allow duplicates */
	while (*mptr != 0) {
		assert(! equalUtf8Consts (name, (*mptr)->name)
		       || ! equalUtf8Consts (signature, (*mptr)->signature));
		mptr = &(*mptr)->next;
	}
#endif

MDBG(	printf("Adding method %s:%s%s (%x)\n", c->name->data, WORD2UTF(pool->data[nc])->data, WORD2UTF(pool->data[sc])->data, m->access_flags);	)

	mt = checked_calloc(sizeof(Method), 1);
	mt->name = name;
	mt->signature = signature;
	mt->class = c;
	mt->accflags = m->access_flags;
	mt->code = 0;
	mt->insn = 0;
	mt->ncode = 0;
	mt->stacksz = 0;
	mt->localsz = 0;
	mt->exception_table = 0;
	mt->exception_table_len = 0;
	mt->exception_next = 0;
	countInsAndOuts(signature->data, &mt->ins, &mt->outs, &mt->outtype);
	mt->next = *mptr;
	*mptr = mt;

	return (mt);
}

fields*
addField(classes* c, field_info* f)
{
	constIndex nc;
	constIndex sc;
	fields** fptr;
	fields* ft;
	constants* pool;
	int* optr;
	char* sig;

	pool = CLASS_CONSTANTS (c);

	nc = f->name_index;
	if (pool->tags[nc] != CONSTANT_Utf8) {
FDBG(		printf("addField: no field name.\n");			)
		return (0);
	}

	/* Search down class for field name */
	if (f->access_flags & ACC_STATIC) {
		fptr = &c->staticFieldList;
		optr = &c->sfsize;
	}
	else {
		fptr = &c->fieldList;
		optr = &c->fsize;
	}
	/* This loop makes field insertion O(N^2).  FIXME. */
	while (*fptr != 0) {
		assert(! equalUtf8Consts (WORD2UTF (pool->data[nc]), 
					 (*fptr)->name));
		fptr = &(*fptr)->next;
	}

FDBG(	printf("Adding field %s:%s\n", c->name, pool->data[nc].v.tstr);	)

	sc = f->signature_index;
	if (pool->tags[sc] != CONSTANT_Utf8) {
FDBG(		printf("addField: no signature name.\n");		)
		return (0);
	}

	ft = checked_calloc(sizeof(Field), 1);
	ft->name = WORD2UTF (pool->data[nc]);
	ft->sig = WORD2UTF (pool->data[sc]);
	ft->accflags = f->access_flags;
	sig = ft->sig->data;
	ft->size = sizeofSig(&sig);
	ft->class = c;
	ft->next = *fptr;
	*fptr = ft;

	return (ft);
}

void
addMethodCode(methods* m, Code* c)
{
	m->code = c->code;
	m->codelen = c->code_length;
	m->stacksz = c->max_stack;
	m->localsz = c->max_locals;
	m->exception_table = c->exception_table;
	m->exception_table_len = c->exception_table_length;
}

void
addInterfaces(classes* c, int inr, classes** inf)
{
	assert(inr > 0);

        c->interfaces = inf;
	c->interface_len = inr;
}

static char loadClassName[] = "loadClass";
static char loadClassSig[] = "(Ljava/lang/String;Z)Ljava/lang/Class;";

/*
 * Lookup a named class, loading it if necessary.
 */
Class*
loadClass(Utf8Const* name, struct _classLoader* loader)
{
	Class* class;

	class = simpleLookupClass(name, loader);
	if (class == NULL) {
		/* Failed to find class, so must now load it */

		if (loader != NULL) {
			stringClass* str;

LDBG(			printf("classLoader: loading %s\n", name->data); )
			str = makeReplaceJavaStringFromUtf8(name->data, name->length, '/', '.');
			class = (classes*)do_execute_java_method (NULL, (object*)loader, loadClassName, loadClassSig, 0, false, str, true);
LDBG(			printf("classLoader: done\n");			)
		}
		else {
			class = findClass(name->data);
		}

		/* Once loaded, allocate the static data space */
		if (class == NULL) {
                        throwException(ClassNotFoundException(name->data));
		}
	}
	prepareClass (class);
	return (class);
}

classes*
lookupClass (char* name)
{
	classes* class;

	class = loadClass (makeUtf8Const (name, strlen (name)), NULL);
	initializeClass (class);
	return class;
}

static
void
resolveFields (classes* class, fields* fld, int *optr)
{
	for (; fld != 0; fld = fld->next) {
		/* Align field offset */
		if (((*optr) % fld->size) == 1)
			(*optr)++;
		fld->offset = *optr;
		(*optr) += fld->size;
	}
}

static
void
prepareClass (classes* class)
{
	methods* meth;
	int i;
	methods** mtab;
	int has_vmethod = 0;
	int super_idx;

	if (class->state >= CSTATE_DOING_PREPARE) {
		if (class->state == CSTATE_DOING_PREPARE) {
			throwException(ClassCircularityError);
		}
		return;
	}
	super_idx = (int) class->superclass;
	class->state = CSTATE_DOING_PREPARE;

	/* Load and link the super class */
	if (super_idx != 0) {
		if (super_idx != -1) {
			class->superclass = getClass (super_idx, class);
		}
		else {
			class->superclass = ObjectClass;
		}
		prepareClass (class->superclass);
		class->msize = class->superclass->msize;
		class->fsize = class->superclass->fsize;
	}

	/* Load all the implemented interfaces. */
	for (i = 0;  i < class->interface_len;  i++) {
		size_t iface = (size_t) class->interfaces[i];
		class->interfaces[i] = getClass (iface, class);
	}

	resolveFields (class, class->fieldList, &class->fsize);
	resolveFields (class, class->staticFieldList, &class->sfsize);

	if (class->sfsize > 0) {
		class->staticFields = (int*)checked_calloc(class->sfsize,sizeof(int));
	}

	/* Build dispatch table. */
	for (meth = class->methodList;  meth != NULL;  meth = meth->next) {
	  	classes* super = class->superclass;
		if (meth->accflags & ACC_STATIC) {
			meth->idx = -1;
			continue;
		}
		has_vmethod++;
		/* Search superclasses for equivalent method name.
		 * If found extract its index nr.
		 */
		for (; super != NULL;  super = super->superclass) {
			methods* mt = super->methodList;
			for (; mt != NULL;  mt = mt->next) {
				if (equalUtf8Consts (mt->name, meth->name)
				    && equalUtf8Consts (mt->signature, meth->signature)) {
					meth->idx = mt->idx;
					goto foundmatch;
				}
			}
		}
		/* No match found so allocate a new index number */
		meth->idx = class->msize++;
	    foundmatch: ;
	}
#if 0
	/* If we change object layout so OBJECT_CLASS
	   does not use dtable, then this makes sense. */
	if (class->superclass != NULL && has_vmethod == 0) {
		/* Reuse superclass's dispatch table. */
		class->dtable = class->superclass->dtable;
		mtab = class->dtable->method;
	}
	else
#endif
	{
		class->dtable = (dispatchTable*)
			checked_calloc(sizeof(dispatchTable)
				       + (class->msize * sizeof(void*)), 1);
		class->dtable->class = class;
		mtab = class->dtable->method;

		/* Install inherited methods into dispatch table. */
		if (class->superclass != NULL) {
			methods** super_mtab
			  = class->superclass->dtable->method;
			for (i = 0;  i < class->superclass->msize;  i++) {
				mtab[i] = super_mtab[i];
			}
		}
		for (meth = class->methodList;  meth != NULL;
		     meth = meth->next) {
			if (meth->idx >= 0) {
				mtab[meth->idx] = meth;
			}
		}
	}

	/* Check for undefined abstract methods if class is not abstract.
	 * See "Java Language Specification" (1996) section 12.3.2. */
	if (! (class->accflags & ACC_ABSTRACT)) {
		for (i = class->msize;  --i >= 0; ) {
			if (mtab[i] == NULL) {
				throwException(AbstractMethodError);
			}
		}
	}

	class->state = CSTATE_PREPARED;
}

void
linkClass (classes* class)
{
	int i, idx;
	constants* pool = CLASS_CONSTANTS (class);
	if (class->state >= CSTATE_LINKED)
		return;

	prepareClass (class);

	/* Second stage class verification - check the class format is okay */
	verify2(class);

	/* Load (only) all classes references in the constant pool.
	 * Replace each CONSTANT_Class by a CONSTANT_ResolvedClass. */
	for (i = 0;  i < pool->size; i++) {
		switch (pool->tags[i]) {
		case CONSTANT_Class:
			pool->data[i] = (ConstSlot)getClass (i, class);
			pool->tags[i] = CONSTANT_ResolvedClass;
			break;
		case CONSTANT_String:
			idx = CLASS_CONST_USHORT1(class, i);
			pool->data[i] = (ConstSlot)Utf8Const2JavaString (WORD2UTF (pool->data[idx]));
			break;
		}
	}

	/* Third stage class verification - check the bytecode is okay */
	verify3 (class);

	prepareClass (class);

	class->state = CSTATE_LINKED;			
}

/* Initializes a class.
 * Does not follow the language spec yet. */

void
initializeClass (Class* class)
{
	methods* meth;
	classes* super;
	if (class->state == CSTATE_OK) {
		return;
	}
	/* Not to spec:  should wait if different thread is doing init. */
	if (class->state == CSTATE_DOINGINIT) {
		return;
	}
	if (class->state < CSTATE_LINKED) {
		linkClass (class);
	}

	super = class->superclass;
	if (super != NULL && super->state != CSTATE_OK) {
		initializeClass (super);
	}
DBG(	printf("Initialising %s static %d\n", class->name, class->sfsize); fflush(stdout); )
	meth = findMethodLocal(class, init_name, void_signature);
	if (meth != NULL) {
		class->state = CSTATE_DOINGINIT;
		CALL_KAFFE_FUNCTION(meth, 0);
	}
	class->state = CSTATE_OK;
}

static
Class*
simpleLookupClass(Utf8Const* name, struct _classLoader* loader)
{
	Class* clp;

	for (clp = classPool[hashClassName(name) & (CLASSHASHSZ-1)]; clp != 0 && !(equalUtf8Consts(name, clp->name) && loader == clp->loader); clp = clp->next)
		;
	return (clp);
}

/*
 * Lookup a named field.
 */
fields*
lookupClassField(classes* clp, Utf8Const* name, bool isStatic)
{
	fields* fptr;

	/* Search down class for field name */
	if (isStatic) {
		fptr = clp->staticFieldList;
	}
	else {
		fptr = clp->fieldList;
	}
	while (fptr != 0) {
		if (equalUtf8Consts (name, fptr->name)) {
			return (fptr);
		}
		fptr = fptr->next;
	}
FDBG(	printf("Class:field lookup failed %s:%s\n", c, f);		)
	return (0);
}

/*
 * Determine the number of arguments and return values from the
 * method signature.
 */
void
countInsAndOuts(char* str, int* ins, int* outs, char* outtype)
{
	*ins = sizeofSig(&str);
	*outtype = str[0];
	*outs = sizeofSig(&str);
}

/*
 * Calculate size of data item based on signature.
 */
int
sizeofSig(char** strp)
{
	int count;
	int c;

	count = 0;
	for (;;) {
		c = sizeofSigItem(strp, 0);
		if (c == -1) {
			return (count);
		}
		count += c;
	}
}

/*
 * Calculate size (in words) of a signature item.
 */
int
sizeofSigItem(char** strp, char* sig)
{
	int count;
	char* str;

	for (str = *strp; ; str++) {

		/* Return sig if requested */
		if (sig != 0) {
			*sig = *str;
		}

		switch (*str) {
		case '(':
			continue;
		case 0:
		case ')':
			count = -1;
			break;
		case 'V':
			count = 0;
			break;
		case 'I':
		case 'Z':
		case 'S':
		case 'B':
		case 'C':
		case 'F':
			count = 1;
			break;
		case 'D':
		case 'J':
			count = 2;
			break;
		case '[':
			count = 1;
			arrayofarray:
			str++;
			if (*str == 'L') {
				while (*str != ';') {
					str++;
				}
			}
			else if (*str == '[') {
				goto arrayofarray;
			}
			break;
		case 'L':
			count = 1;
			/* Skip to end of reference */
			while (*str != ';') {
				str++;
			}
			break;
		default:
			abort();
		}

		*strp = str + 1;
		return (count);
	}
}

/*
 * Add this method to the method list so we can find it when
 * handling exceptions.
 */
void
establishMethod(methods* m)
{
	m->exception_next = methodList;
	methodList = m;
}

/*
 * Lookup an array class, using its name (signature).
 */
classes*
lookupArray(Utf8Const* name)
{
	classes* class;

	class = simpleLookupClass(name, NULL);
	if (class == 0) {
		class = internalAddClass (alloc_class (), name, 0, 0, NULL);
CDBG(		if (class == 0) {
			printf("Array Class %s not loaded.\n", name->data);
		}							)
		class->superclass = (Class*)-1;
		prepareClass (class);
		class->state = CSTATE_OK;
	}
	return (class);
}

/* Find (or create) an array class with component type C. */

Class*
lookupObjectArrayClass(Class* c)
{
	char* cname = CLASS_CNAME (c);
        Utf8Const *arr_name;
        char sig[CLASSMAXSIG];  /* FIXME! unchecked fixed buffer! */
        classes* arr_class;

        /* Build signature for array type */
        sprintf (sig, cname[0] == '[' ? "[%s" : "[L%s;", cname);
        arr_name = makeUtf8Const (sig, strlen (sig));
        arr_class = simpleLookupClass(arr_name, c->loader);
        if (arr_class == NULL) {
                arr_class = internalAddClass (alloc_class (),
					      arr_name, 0, 0, c->loader);
                arr_class->superclass = (Class*)-1;
                prepareClass (arr_class);
                arr_class->state = CSTATE_OK;
        }
#if !INTERN_UTF8CONSTS
	else {
		free (arr_name);
	}
#endif
	return (arr_class);
}

/*
 * Load a class or array depending on the signature and return the
 * class object.
 */
Class*
loadClassOrArray(char* data, uint16 len, struct _classLoader* loader)
{
	switch (data[0]) {
	case '[':
		return (lookupArray(makeUtf8Const(data, len)));
	case 'L':
		assert(data[len-1] == ';');
		return (loadClass(makeUtf8Const(data+1, len-2), loader));
	default:
		return (0);
	}
}
