/*
 * classMethod.h
 * Class, method and field tables.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#ifndef __classmethod_h
#define __classmethod_h

#include "gtypes.h"
#include "object.h"
#include "constants.h"

#define	MAXMETHOD		64

/* Class state */
#define	CSTATE_LOADED		0
#define CSTATE_DOING_PREPARE	1
#define CSTATE_PREPARED		2
#define CSTATE_LINKED		3
#define	CSTATE_DOINGINIT	4
#define	CSTATE_OK		5

typedef struct _classLoader {
	int	dummy;
} classLoader;

typedef struct _classes {
	Object			head;		/* A class is an object too */
	Class*			next;
	Utf8Const*		name;
	int			accflags;

	/* If non-NULL, a pointer to the superclass.
	 * However, if state < CSTATE_DOING_PREPARE, then
	 * (int) superclass is a constant pool index. */
	Class*			superclass;

	struct _constants	constants;
	struct _methods*	methodList;

	/* Number of methods in the dtable. */
	int			msize;

	struct _fields*		fieldList;
	int			fsize;
	struct _fields*		staticFieldList;

	/* Size (in words) of the static fields. */
	int			sfsize;
	struct _dispatchTable*	dtable;
	int*			staticFields;
        struct _classes**       interfaces;
	int			interface_len;
	struct _classLoader*	loader;
	int			state;
	bool			final;
} classes;

typedef struct _methods {
	Method*			next;
	Utf8Const*		name;
	Utf8Const*		signature;
	accessFlags		accflags;
	char*			code;
	int			codelen;
	struct _instn*		insn;
	nativecode*		ncode;
	nativecode*		ncode_end;
	struct _jexception*	exception_table;
	int			exception_table_len;
	Method*			exception_next;
	int			stacksz;
	int			localsz;
	int			ins;
	int			outs;
	char			outtype;
	Class*			class;

	/* Index into class->dtable.  Is -1 if method is static. */
	int			idx;
} methods;

typedef struct _dispatchTable {
	classes*		class;
	methods*		method[1];
} dispatchTable;

#define	CLASS_STATE		((int)&(((classes*)0)->state))

#define	DTABLE_CLASS		0
#define	DTABLE_METHODOFFSET	4
#define	DTABLE_METHODSIZE	(sizeof(void*))

typedef struct _fields {
	Field*			next;
	Utf8Const*		name;
	Utf8Const*		sig;
	accessFlags		accflags;
	int			size; /* in words */
	uintp			offset; /* in words. */
	classes*		class; /* The class I'm in */
} fields;

#define FIELD_WSIZE(FLD)	((FLD)->size)
#define FIELD_SIZE(FLD)		(4 * FIELD_WSIZE(FLD))
#define FIELD_WOFFSET(FLD)	((FLD)->offset)
#define FIELD_OFFSET(FLD)	(4 * (FLD)->offset)
#define FIELD_ISREF(FLD)	\
	((FLD)->sig->data[0] == 'L' || (FLD)->sif->data[0] == '[')

#define	CLASSMAXSIG		256

struct _Code;
struct _method_info;
struct _field_info;
struct _classFile;

#define OBJECT_CLASS(OBJ)		((OBJ)->dtable->class)
#define	METHOD_CLASS(METH)		((METH)->class)
#define	METHOD_CONST(METH)		(&((METH)->class->constants))
#define CLASS_CNAME(CL)			((CL)->name->data)

classes*	addClass(Class*, constIndex, constIndex, u2, struct _classLoader*);
methods*	addMethod(classes*, struct _method_info*);
void 		addMethodCode(methods*, struct _Code*);
fields*         addField(classes*, struct _field_info*);
void		addCode(methods*, uint32, struct _classFile*);
classes*	loadClass(Utf8Const*, struct _classLoader*);
classes*	loadClassOrArray(char*, uint16, struct _classLoader*);
void		initializeClass (classes*);
extern classes*	lookupClass(char*);
classes*	lookupArray(Utf8Const*);
Class*		lookupObjectArrayClass(Class*);
fields*		lookupClassField(classes*, Utf8Const*, bool);
void		countInsAndOuts(char*, int*, int*, char*);
int		sizeofSig(char**);
int		sizeofSigItem(char**, char*);
void		establishMethod(methods*);
void		addInterfaces(classes*, int, classes**);
extern classes* getClass(constIndex, classes*);
extern Utf8Const* makeUtf8Const (char*, int);
extern struct _stringClass* Utf8Const2JavaString (Utf8Const*);

extern Utf8Const* init_name;		/* "<clinit>" */
extern Utf8Const* constructor_name;	/* "<init>" */
extern Utf8Const* final_name;		/* "finalize" */
extern Utf8Const* void_signature;	/* "()V" */
extern Utf8Const* Code_name;	/* "Code" */

#define VOID_SIGNATURE (void_signature->data)  /* same as "()V" but shared */

#endif
