/*
 * lookup.c
 * Various lookup calls for resolving objects, methods, exceptions, etc.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define DBG(s)
#define FDBG(s)
#define EDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "errors.h"
#include "slots.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "lookup.h"
#include "exception.h"
#include "thread.h"
#include "baseClasses.h"
#include "machine.h"

extern methods* methodList;
extern classes* ObjectClass;

/*
 * Lookup a method reference and get the various interesting bits.
 */
void
getMethodSignatureClass(constIndex idx, methods* meth, callInfo* call)
{
	constants* pool = CLASS_CONSTANTS (meth->class);
	constIndex ci;
	constIndex ni;
	constIndex nin;
	constIndex nis;
	classes* class;
	Utf8Const* name;
	Utf8Const* sig;
	char* s;
	int i;
	int r;
        methods* mptr;
        classes* cptr;

	if (pool->tags[idx] != CONSTANT_Methodref &&
	    pool->tags[idx] != CONSTANT_InterfaceMethodref) {
DBG(		printf("No Methodref found\n");				)
                throwException(NoSuchMethodError);
	}

	ni = METHODREF_NAMEANDTYPE(idx, pool);
	nin = NAMEANDTYPE_NAME(ni, pool);
	name = WORD2UTF(pool->data[nin]);
	nis = NAMEANDTYPE_SIGNATURE(ni, pool);
	sig = WORD2UTF(pool->data[nis]);

	ci = METHODREF_CLASS(idx, pool);
	class = CLASS_CLASS (ci, pool);

DBG(	printf("getMethodSignatureClass(%s,%s,%s)\n",
		class->name, pool->data[nin], pool->data[nis]);	)

	/*
	 * We now have a complete class and method signature.  We can
	 * Now work out some things we might need to know, like no. of
	 * arguments, address of function, etc.
	 */
	countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
DBG(	printf("in = %d, out = %d\n", call->in, call->out);		)
	call->class = class;

	/* Find method */
        for (cptr = class; cptr != 0; cptr = cptr->superclass) {
                for (mptr = cptr->methodList; mptr != 0; mptr = mptr->next) {
			if (equalUtf8Consts (name, mptr->name)
			    && equalUtf8Consts (sig, mptr->signature)) {
				goto foundit;
			}
		}
	}
	assert("Method not found" == 0);
	return;

	foundit:
	call->method = mptr;
}

classes*
getClass(constIndex idx, classes* this)
{
	constants* pool;
	constIndex cs;
	Utf8Const *name;
	classes* class;

	pool = CLASS_CONSTANTS(this);
	if (pool->tags[idx] == CONSTANT_ResolvedClass) {
		return (CLASS_CLASS (idx, pool));
	}
	if (pool->tags[idx] != CONSTANT_Class) {
		throwException(ClassFormatError);
	}
	cs = CLASS_NAME(idx, pool);
	name = WORD2UTF(pool->data[cs]);
	/* Find the specified class.  We cannot use 'loadClassOrArray' here
	 * because the name is *not* a signature.
	 */
	if (name->data[0] == '[') {
		class = lookupArray(name);
	}
	else {
		class = loadClass(name, this->loader);
	}
	if (class == 0) {
		throwException(ClassNotFoundException(name->data));
	}
	return (class);
}

Field*
getField(constIndex idx, bool isStatic, Method* meth)
{
	constants* pool = CLASS_CONSTANTS (meth->class);
	constIndex ci;
	constIndex ni;
	constIndex nin;
	constIndex nis;
	Field* field;
	char* sig;
	classes* class;

	if (pool->tags[idx] == CONSTANT_ResolvedFieldref) {
		return (CLASS_FIELD (idx, pool));
	}
	if (pool->tags[idx] != CONSTANT_Fieldref) {
FDBG(		printf("No Fieldref found\n");				)
                throwException(NoSuchFieldError);
	}

	ci = FIELDREF_CLASS(idx, pool);
	ni = FIELDREF_NAMEANDTYPE(idx, pool);
	nin = NAMEANDTYPE_NAME(ni, pool);
	nis = NAMEANDTYPE_SIGNATURE(ni, pool);
	class = CLASS_CLASS (ci, pool);

FDBG(	printf("*** getField(%s,%s,%s)\n",
		class->name, pool->data[nin], pool->data[nis]);	)

	field = lookupClassField(class, WORD2UTF(pool->data[nin]),
				 isStatic);
	if (field == 0) {
FDBG(		printf("Field not found\n");				)
                throwException(NoSuchFieldError);
	}

	/* Rewrite field as resolved */
	pool->tags[idx] = CONSTANT_ResolvedFieldref;
	pool->data[idx] = (ConstSlot)field;

	return (field);
}

/*
 * Lookup a method (and translate) in the specified class.
 */
methods*
findMethodLocal (classes* class, Utf8Const* name, Utf8Const* signature)
{
	methods* mptr;
	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	for (mptr = class->methodList; mptr != 0; mptr = mptr->next) {
		if (equalUtf8Consts (name, mptr->name)
		    && equalUtf8Consts (signature, mptr->signature)) {
#if defined(TRANSLATOR)
			/*
			 * Generate code on demand.
			 */
			if (mptr->ncode == 0) {
				translate(mptr);
			}
#endif
			break;
		}
	}
	return (mptr);
}

/*
 * Lookup a method (and translate) in the specified class.
 */
methods*
findMethod(classes* class, Utf8Const* name, Utf8Const* signature)
{
	linkClass (class);
	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	for (; class != 0; class = class->superclass) {
		methods* mptr = findMethodLocal (class, name, signature);
		if (mptr != NULL) {
			return mptr;
		}
	}
	return (0);
}

#if defined(TRANSLATOR)
/*
 * Find exception in method.
 */
void
findExceptionInMethod(uintp pc, classes* class, exceptionInfo* info)
{
	methods* ptr;

	info->handler = 0;
	info->class = 0;
	info->method = 0;

	for (ptr = methodList; ptr != 0; ptr = ptr->exception_next) {

		if (pc < (uintp)ptr->ncode || pc >= (uintp)ptr->ncode_end) {
			continue;
		}
EDBG(		printf("Found method %s.%s%s\n", ptr->class->name,
			ptr->pair->s1, ptr->pair->s2);			)

		if (findExceptionBlockInMethod(pc, class, ptr, info) == true) {
			return;
		}
	}
EDBG(	printf("Exception not found.\n");				)
}

/*
 * Find method containing pc.
 */
methods*
findMethodFromPC(uintp pc)
{
	methods* ptr;

	for (ptr = methodList; ptr != 0; ptr = ptr->exception_next) {
		if (pc >= (uintp)ptr->ncode && pc < (uintp)ptr->ncode_end) {
EDBG(			printf("Found method %s.%s%s\n",
				ptr->class->name,
				ptr->pair->s1, ptr->pair->s2);		)
			break;
		}
	}
EDBG(	if (ptr == 0) printf("Method not found.\n");			)

	return (ptr);
}
#endif

/*
 * Look for exception block in method.
 */
bool
findExceptionBlockInMethod(uintp pc, classes* class, methods* ptr, exceptionInfo* info)
{
	jexception* eptr;
	classes* cptr;
	int i;

	/* Stash method & class */
	info->method = ptr;
	info->class = ptr->class;

	eptr = ptr->exception_table;

EDBG(	printf("Nr of exceptions = %d\n", ptr->exception_table_len); )

	/* Right method - look for exception */
	for (i = 0; i < ptr->exception_table_len; i++) {
EDBG(		printf("Exceptions %x (%x-%x)\n", pc, eptr[i].start_pc, eptr[i].end_pc); )
		if (pc < eptr[i].start_pc || pc > eptr[i].end_pc) {
			continue;
		}
EDBG(		printf("Found exception 0x%x\n", eptr[i].handler_pc); )

		/* Found exception - is it right type */
		if (eptr[i].catch_type == NULL) {
			info->handler = eptr[i].handler_pc;
			return (true);
		}
		for (cptr = class; cptr != 0; cptr = cptr->superclass) {
			if (cptr == eptr[i].catch_type) {
				info->handler = eptr[i].handler_pc;
				return (true);
			}
		}
	}
	return (false);
}
