/*
 * thread.c
 * Thread support.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define	DBG(s)
#define	SDBG(s) 

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-signal.h"
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "thread.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "errors.h"
#include "md.h"

#if defined(USE_INTERNAL_THREADS)

#if defined(INTERPRETER)
extern vmException* cjbuf;
#endif

thread* currentThread;
thread* threadQhead[MAX_THREAD_PRIO + 1];
thread* threadQtail[MAX_THREAD_PRIO + 1];

thread* garbageman;
thread* finalman;

thread* liveThreads;
thread* alarmList;
classes* ThreadClass;

int blockInts;
bool needReschedule;

static int talive;
static int tdaemon;
static bool alarmBlocked;

static void firstStartThread(void);
static thread* startDaemon(void*, char*);
static void addToAlarmQ(thread*, jlong);
static void alarmException(int);

/* Select an alarm system */
#if defined(HAVE_SETITIMER) && defined(ITIMER_REAL)
#define	MALARM(_mt)							\
	{								\
		struct itimerval tm;					\
		tm.it_interval.tv_sec = 0;				\
		tm.it_interval.tv_usec = 0;				\
		tm.it_value.tv_sec = (_mt) / 1000;			\
		tm.it_value.tv_usec = ((_mt) % 1000) * 1000;		\
		setitimer(ITIMER_REAL, &tm, 0);				\
	}
#elif defined(HAVE_ALARM)
#define	MALARM(_mt)	alarm((int)(((_mt) + 999) / 1000))
#endif

void reschedule(void);
void finaliserMan(void);
void gcMan(void);
void checkEvents(bool);

/* Setup default thread stack size - this can be overwritten if required */
int threadStackSize = THREADSTACKSIZE;

#define	ALLOCTHREADSTACK(_t)						\
	(_t)->PrivateInfo->stackBase = checked_malloc(threadStackSize);	\
	assert((_t)->PrivateInfo->stackBase != 0);			\
	(_t)->PrivateInfo->stackEnd = (_t)->PrivateInfo->stackBase + 	\
		threadStackSize

#define	FREETHREADSTACK(_t)						\
	if (!((_t)->PrivateInfo->flags & THREAD_FLAGS_NOSTACKALLOC)) {	\
		checked_free((_t)->PrivateInfo->stackBase);		\
	}								\
	(_t)->PrivateInfo->stackBase = 0;				\
	(_t)->PrivateInfo->stackEnd = 0



/*
 * Initialise threads.
 */
void
initThreads(void)
{
	/* Get a handle on the thread class */
	ThreadClass = lookupClass(THREADCLASS);
	assert(ThreadClass != 0);

	/* Allocate a thread to be the main thread */
	currentThread = (thread*)alloc_object(ThreadClass, false);
	assert(currentThread != 0);
	currentThread->PrivateInfo = checked_calloc(sizeof(ctx), 1);
	liveThreads = currentThread;

	currentThread->name = makeJavaCharArray("main", strlen("main"));
	currentThread->priority = NORM_THREAD_PRIO;
	currentThread->PrivateInfo->priority = (uint8)currentThread->priority;
	currentThread->next = 0;
	currentThread->PrivateInfo->status = THREAD_SUSPENDED;
	assert(currentThread->PrivateInfo != 0);
	THREADINFO(currentThread->PrivateInfo);
DBG(	printf("main thread %x base %x end %x\n", currentThread, currentThread->PrivateInfo->stackBase, currentThread->PrivateInfo->stackEnd); )
	currentThread->PrivateInfo->flags = THREAD_FLAGS_NOSTACKALLOC;
	currentThread->PrivateInfo->nextlive = 0;
	currentThread->PrivateInfo->nextalarm = 0;
	currentThread->PrivateInfo->blockqueue = 0;
	currentThread->single_step = 0;
	currentThread->daemon = 0;
	currentThread->stillborn = 0;
	currentThread->target = 0;
	currentThread->interruptRequested = 0;
	currentThread->group = (threadGroup*)execute_java_constructor(0, THREADGROUPCLASS, 0, VOID_SIGNATURE);
	assert(currentThread->group != 0);

	talive++;

	/* Add thread into runQ */
	iresumeThread(currentThread);

	/* Start any daemons we need */
	finalman = startDaemon(&finaliserMan, "finaliser");
	garbageman = startDaemon(&gcMan, "gc");
	iresumeThread(finalman);
	iresumeThread(garbageman);

	/* Plug in the alarm handler */
#if defined(SIGALRM)
	signal(SIGALRM, (SIG_T)alarmException);
#else
#error "No alarm signal support"
#endif
}

/*
 * Start a new thread running.
 */
void
startThread(thread* tid)
{
	/* Allocate a stack context */
	assert(tid->PrivateInfo == 0);
	tid->PrivateInfo = checked_calloc(sizeof(ctx), 1);
	tid->PrivateInfo->nextlive = liveThreads;
	tid->PrivateInfo->nextalarm = 0;
	tid->PrivateInfo->blockqueue = 0;
	liveThreads = tid;
	ALLOCTHREADSTACK(tid);
	tid->PrivateInfo->flags = THREAD_FLAGS_GENERAL;
	tid->PrivateInfo->status = THREAD_SUSPENDED;
	tid->PrivateInfo->priority = (uint8)tid->priority;

	/* Construct the initial restore point. */
	THREADINIT(tid->PrivateInfo, &firstStartThread);
DBG(	printf("new thread %x base %x end %x\n", tid, tid->PrivateInfo->stackBase, tid->PrivateInfo->stackEnd); )

	talive++;
	if (tid->daemon) {
		tdaemon++;
	}

	/* Add thread into runQ */
	iresumeThread(tid);
}

/*
 * All threads start here.
 */
static
void
firstStartThread(void)
{
DBG(	printf("firstStartThread %x\n", currentThread);		)
	/* Every thread starts with the interrupts off */
	intsRestore();

	/* Find the run()V method and call it */
	do_execute_java_method(0, (object*)currentThread, "run", "()V", 0, 0);
	do_execute_java_method(0, (object*)currentThread, "exit", "()V", 0, 0);
	killThread(0);
	assert("Thread returned from killThread" == 0);
}


/*
 * Explicit request by user to resume a thread
 * The definition says that it is just legal to call this after a preceeding
 * suspend (which got through). If the thread was blocked for some other
 * reason (either sleep or IO or a muxSem), we simply can't do it
 * We use a new thread flag THREAD_USERSUSPEND for this purpose
 * (which is set by suspendThread(.))
 */
void
resumeThread(thread* tid)
{
        if ((tid->PrivateInfo->status & THREAD_USERSUSPEND) != 0) {
                intsDisable();
                tid->PrivateInfo->status &= ~THREAD_USERSUSPEND;
                iresumeThread(tid);
                intsRestore();
        }
}

/*
 * Resume a thread running.
 * This routine has to be called only from locations which ensure
 * run / block queue consistency. There is no check for illegal resume
 * conditions (like explicitly resuming an IO blocked thread). There also
 * is no update of any blocking queue. Both has to be done by the caller
 */
void
iresumeThread(thread* tid)
{
DBG(	printf("resumeThread %x\n", tid);			)
	intsDisable();

	if (tid->PrivateInfo->status != THREAD_RUNNING) {
		tid->PrivateInfo->status = THREAD_RUNNING;

		/* Place thread on the end of its queue */
		if (threadQhead[tid->PrivateInfo->priority] == 0) {
			threadQhead[tid->PrivateInfo->priority] = tid;
			threadQtail[tid->PrivateInfo->priority] = tid;
			if (tid->PrivateInfo->priority > currentThread->PrivateInfo->priority) {
				needReschedule = true;
			}
		}
		else {
			threadQtail[tid->PrivateInfo->priority]->next = tid;
			threadQtail[tid->PrivateInfo->priority] = tid;
		}
		tid->next = 0;
	}
SDBG(	else {
		printf("Re-resuming 0x%x\n", tid);
	}							)
	intsRestore();
}

/*
 * Yield process to another thread of equal priority.
 */
void
yieldThread()
{
	intsDisable();

	if (threadQhead[currentThread->PrivateInfo->priority] != threadQtail[currentThread->PrivateInfo->priority]) {

		/* Get the next thread and move me to the end */
		threadQhead[currentThread->PrivateInfo->priority] = currentThread->next;
		threadQtail[currentThread->PrivateInfo->priority]->next = currentThread;
		threadQtail[currentThread->PrivateInfo->priority] = currentThread;
		currentThread->next = 0;
		needReschedule = true;
	}

	intsRestore();
}

/*
 * Suspend a thread.
 * This is an explicit user request to suspend the thread - the counterpart
 * for resumeThreadRequest(.). It is JUST called by the java method
 * Thread.suspend()
 * What makes it distinct is the fact that the suspended thread is not contained
 * in any block queue. Without a special flag (indicating the user suspend), we
 * can't check s suspended thread for this condition afterwards (which is
 * required by resumeThreadRequest()). The new thread flag
 * THREAD_USERSUSPEND is used for this purpose.
 */
void
suspendThread(thread* tid)
{
	thread** ntid;

	intsDisable();

	if (tid->PrivateInfo->status != THREAD_SUSPENDED) {
		tid->PrivateInfo->status = THREAD_SUSPENDED;
		
		/*
                 * This is used to indicate the explicit suspend condition
                 * required by resumeThreadRequest()
                 */
                tid->PrivateInfo->status |= THREAD_USERSUSPEND;

		for (ntid = &threadQhead[tid->PrivateInfo->priority]; *ntid != 0; ntid = &(*ntid)->next) {
			if (*ntid == tid) {
				*ntid = tid->next;
				tid->next = 0;
				if (tid == currentThread) {
					reschedule();
				}
				break;
			}
		}
	}
SDBG(	else {
		printf("Re-suspending 0x%x\n", tid);
	}							)

	intsRestore();
}

/*
 * Suspend a thread on a queue.
 */
void
suspendOnQThread(thread* tid, thread** queue, jlong timeout)
{
	thread** ntid;
	thread* last;

DBG(	printf("suspendOnQThread %x %x (%d)\n", tid, queue, (int)timeout); )

	assert(blockInts == 1);

	if ((tid->PrivateInfo->status & THREAD_SUSPENDED) == 0) {
		tid->PrivateInfo->status = THREAD_SUSPENDED;

		last = 0;
		for (ntid = &threadQhead[tid->PrivateInfo->priority]; *ntid != 0; ntid = &(*ntid)->next) {
			if (*ntid == tid) {

				/* Remove thread from runq */
				*ntid = tid->next;
				if (*ntid == 0) {
					threadQtail[tid->PrivateInfo->priority] = last;
				}

				/* Insert onto head of lock wait Q */
				if (queue != 0) {
					tid->next = *queue;
					*queue = tid;
					tid->PrivateInfo->blockqueue = queue;
				}

				/* If I have a timeout, insert into alarmq */
				if (timeout >= 0) {
					addToAlarmQ(tid, timeout);
				}

				/* If I was running, reschedule */
				if (tid == currentThread) {
					reschedule();
				}
				break;
			}
			last = *ntid;
		}
	}
SDBG(	else {
		printf("Re-suspending 0x%x on %x\n", tid, *queue);
	}							)
}

/*
 * Kill thread.
 */
void
killThread(thread* tid)
{
	thread** ntid;

	intsDisable();

	/* A null tid means the current thread */
	if (tid == 0) {
		tid = currentThread;
	}

DBG(	printf("killThread %x\n", tid);			)

	if (tid->PrivateInfo->status != THREAD_DEAD) {

		/* Get thread off runq (if it needs it) */
		if (tid->PrivateInfo->status == THREAD_RUNNING) {
			for (ntid = &threadQhead[tid->PrivateInfo->priority]; *ntid != 0; ntid = &(*ntid)->next) {
				if (*ntid == tid) {
					*ntid = tid->next;
					break;
				}
			}
		}

		tid->PrivateInfo->status = THREAD_DEAD;
		talive--;
		if (tid->daemon) {
			tdaemon--;
		}

		/* If we only have daemons left, then everyone is dead. */
		if (talive == tdaemon) {
			/* Am I suppose to close things down nicely ?? */
			exit(0);
		}

		/* Notify on the object just in case anyone is waiting */
		lockMutex(&tid->obj);
		broadcastCond(&tid->obj); 
		unlockMutex(&tid->obj);

		/* Remove thread from live list to it can be garbaged */
		for (ntid = &liveThreads; *ntid != 0; ntid = &(*ntid)->PrivateInfo->nextlive) {
			if (tid == (*ntid)) {
				(*ntid) = tid->PrivateInfo->nextlive;
				break;
			}
		}

		/* Free stack */
		FREETHREADSTACK(tid);

		/* Run something else */
		needReschedule = true;
	}
	intsRestore();
}

/*
 * Change thread priority.
 */
void
setPriorityThread(thread* tid, int prio)
{
	thread** ntid;
	thread* last;

	if (tid->PrivateInfo == 0) {
		tid->priority = prio;
		return;
	}

	if ((tid->PrivateInfo->status & THREAD_SUSPENDED) != 0) {
		tid->PrivateInfo->priority = (uint8)prio;
		return;
	}

	intsDisable();

	/* Remove from current thread list */
	last = 0;
	for (ntid = &threadQhead[tid->PrivateInfo->priority]; *ntid != 0; ntid = &(*ntid)->next) {
		if (*ntid == tid) {
			*ntid = tid->next;
			if (*ntid == 0) {
				threadQtail[tid->PrivateInfo->priority] = last;
			}
			break;
		}
		last = *ntid;
	}

	/* Insert onto a new one */
	tid->priority = prio;
	tid->PrivateInfo->priority = (uint8)tid->priority;
	if (threadQhead[prio] == 0) {
		threadQhead[prio] = tid;
		threadQtail[prio] = tid;
		if (prio > currentThread->PrivateInfo->priority) {
			needReschedule = true;
		}
	}
	else {
		threadQtail[prio]->next = tid;
		threadQtail[prio] = tid;
	}
	tid->next = 0;

	intsRestore();
}

/*
 * Put a thread to sleep.
 */
void
sleepThread(int64 time)
{
	thread** tidp;

	/* Sleep for no time */
	if (time == 0) {
		return;
	}

	intsDisable();

	/* Suspend thread on it */
	suspendOnQThread(currentThread, 0, time);

	intsRestore();
}

static
void
addToAlarmQ(thread* tid, jlong timeout)
{
	thread** tidp;

	assert(blockInts == 1);

	tid->PrivateInfo->status |= THREAD_ALARM;

	/* Get absolute time */
	tid->PrivateInfo->time = timeout + currentTime();

	/* Find place in alarm list and insert it */
	for (tidp = &alarmList; (*tidp) != 0; tidp = &(*tidp)->PrivateInfo->nextalarm) {
		if ((*tidp)->PrivateInfo->time > tid->PrivateInfo->time) {
			break;
		}
	}
	tid->PrivateInfo->nextalarm = *tidp;
	*tidp = tid;

	/* If I'm head of alarm list, restart alarm */
	if (tidp == &alarmList) {
		MALARM(timeout);
	}
}

/*
 * Handle alarm.
 * This routine uses a different meaning of "blockInts". Formerly, it was just
 * "don't reschedule if you don't have to". Now it is "don't do ANY
 * rescheduling actions due to an expired timer". An alternative would be to
 * block SIGALARM during critical sections (by means of sigprocmask). But
 * this would be required quite often (for every outmost intsDisable(),
 * intsRestore()) and therefore would be much more expensive than just
 * setting an int flag which - sometimes - might cause an additional
 * setitimer call.
 */
static
void
alarmException(int sig)
{
	thread* tid;
	thread** ntid;
	int64 time;

	/* Re-enable signal - necessary for SysV */
	signal(sig, (SIG_T)alarmException);

	/*
	 * If ints are blocked, this might indicate an inconsistent state of
	 * one of the thread queues (either alarmList or threadQhead/tail).
	 * We better don't touch one of them in this case and come back later.
	 */
	if (blockInts > 0) {
		MALARM(50);
		return;
	}

	intsDisable();

	/* Wake all the threads which need waking */
	time = currentTime();
	while (alarmList != 0 && alarmList->PrivateInfo->time <= time) {

		/* Remove thread from alarmq */
		tid = alarmList;
		alarmList = alarmList->PrivateInfo->nextalarm;

		/* If blocked on a queue, remove it */
		if (tid->PrivateInfo->blockqueue != 0) {
			for (ntid = tid->PrivateInfo->blockqueue; *ntid != 0; ntid = &(*ntid)->next) {
				if (*ntid == tid) {
					*ntid = tid->next;
					break;
				}
			}
			tid->PrivateInfo->blockqueue = 0;
			tid->interruptRequested = 1;
		}

		/* Restart thread */
		iresumeThread(tid);
	}

	/* Restart alarm */
	if (alarmList != 0) {
		MALARM(alarmList->PrivateInfo->time - time);
	}

	/*
	 * The next bit is rather tricky.  If we don't reschedule then things
	 * are fine, we exit this handler and everything continues correctly.
	 * On the otherhand, if we do reschedule, we will schedule the new
	 * thread with alarms blocked which is wrong.  However, we cannot
	 * unblock them here incase we have just set an alarm which goes
	 * off before the reschedule takes place (and we enter this routine
	 * recusively which isn't good).  So, we set a flag indicating alarms
	 * are blocked, and allow the rescheduler to unblock the alarm signal
	 * after the context switch has been made.  At this point it's safe.
	 */
	alarmBlocked = true;
	intsRestore();
	alarmBlocked = false;
}

/*
 * Is this thread alive?
 */
bool
aliveThread(thread* tid)
{
	if (tid->PrivateInfo != 0 && tid->PrivateInfo->status != THREAD_DEAD) {
		return (true);
	}
	else {
		return (false);
	}
}

/*
 * How many stack frames have I invoked?
 */
long
framesThread(thread* tid)
{
	long count;
	THREADFRAMES(tid, count);
	return (count);
}

/*
 * Reschedule the thread.
 * Called whenever a change in the running thread is required.
 */
void
reschedule(void)
{
	int i;
	thread* lastThread;
	int b;
	sigset_t nsig;

	/* A reschedule in a non-blocked context is half way to hell */
	assert(blockInts > 0);
	b = blockInts;

	/* Check events - we may release a high priority thread */
	/* Just check IO, no need for a reschedule call by checkEvents() */
        needReschedule = false;
	checkEvents(false);

	for (;;) {
	        for (i = MAX_THREAD_PRIO; i >= MIN_THREAD_PRIO; i--) {
			if (threadQhead[i] != 0) {
				if (threadQhead[i] != currentThread) {
#if defined(INTERPRETER)
					currentThread->PrivateInfo->exceptPtr = (void*)cjbuf;
					cjbuf = 0;
#endif
					lastThread = currentThread;
					currentThread = threadQhead[i];

					THREADSWITCH(currentThread->PrivateInfo, lastThread->PrivateInfo);
#if defined(INTERPRETER)
					cjbuf = (vmException*)lastThread->PrivateInfo->exceptPtr;
#endif
                                        blockInts = b;

					/* Alarm signal may be blocked - if so
					 * unblock it.
					 */
					if (alarmBlocked == true) {
						alarmBlocked = false;
						sigemptyset(&nsig);
						sigaddset(&nsig, SIGALRM);
						sigprocmask(SIG_UNBLOCK, &nsig, 0);
					}

					/* I might be dying */
					if ((lastThread->PrivateInfo->flags & THREAD_FLAGS_KILLED) != 0) {
						lastThread->PrivateInfo->flags &= ~THREAD_FLAGS_KILLED;
						throwException(ThreadDeath);
					}
				}
				/* Now we kill the schedule and turn ints
				   back on */
				needReschedule = false;
				return;
			}
		}
		/* Nothing to run - wait for external event */
		checkEvents(true);
	}
}

/*
 * Start a daemon thread.
 */
static
thread*
startDaemon(void* func, char* nm)
{
	thread* tid;

DBG(	printf("startDaemon %s\n", nm);				)

	/* Keep daemon threads as root objects */
	tid = (thread*)alloc_object(ThreadClass, true);
	assert(tid != 0);
	tid->PrivateInfo = checked_calloc(sizeof(ctx), 1);

	tid->name = makeJavaCharArray(nm, strlen(nm));
	tid->priority = MAX_THREAD_PRIO;
	tid->PrivateInfo->priority = (uint8)tid->priority;
	tid->next = 0;
	tid->PrivateInfo->status = THREAD_SUSPENDED;
	assert(tid->PrivateInfo != 0);

	/* Insert into live list as long as it's not the finaliser */
	if (func != &finaliserMan) {
		tid->PrivateInfo->nextlive = liveThreads;
		liveThreads = tid;
	}

	ALLOCTHREADSTACK(tid);
	tid->single_step = 0;
	tid->daemon = 1;
	tid->stillborn = 0;
	tid->target = 0;
	tid->interruptRequested = 0;
	tid->group = 0;

	/* Construct the initial restore point. */
	THREADINIT(tid->PrivateInfo, func);

	talive++;
	tdaemon++;

	return (tid);
}

#endif
